using DataStructures

"""
    gen_code(graph::DAG)

Generate the code for a given graph. The return value is a named tuple of:

- `code::Expr`: The julia expression containing the code for the whole graph.
- `inputSymbols::Dict{String, Vector{Symbol}}`: A dictionary of symbols mapping the names of the input nodes of the graph to the symbols their inputs should be provided on.
- `outputSymbol::Symbol`: The symbol of the final calculated value

See also: [`execute`](@ref)
"""
function gen_code(graph::DAG, machine::Machine)
    nodeQueue = PriorityQueue{Node, Int}()
    inputSyms = Dict{String, Vector{Symbol}}()

    # use a priority equal to the number of unseen children -> 0 are nodes that can be added
    for node in get_entry_nodes(graph)
        enqueue!(nodeQueue, node => 0)
        if !haskey(inputSyms, node.name)
            inputSyms[node.name] = Vector{Symbol}()
        end

        push!(inputSyms[node.name], Symbol("$(to_var_name(node.id))_in"))
    end

    schedule = Vector{Node}()
    sizehint!(schedule, length(graph.nodes))

    # "scheduling"
    node = nothing
    while !isempty(nodeQueue)
        @assert peek(nodeQueue)[2] == 0
        node = dequeue!(nodeQueue)

        push!(schedule, node)
        for parent in node.parents
            # reduce the priority of all parents by one
            if (!haskey(nodeQueue, parent))
                enqueue!(nodeQueue, parent => length(parent.children) - 1)
            else
                nodeQueue[parent] = nodeQueue[parent] - 1
            end
        end
    end

    codeAcc = Vector{Expr}()
    sizehint!(codeAcc, length(graph.nodes))

    for node in schedule
        push!(codeAcc, get_expression(node, machine.devices[1]))
    end

    # node is now the last node we looked at -> the output node
    outSym = Symbol(to_var_name(node.id))

    return (code = Expr(:block, codeAcc...), inputSymbols = inputSyms, outputSymbol = outSym)
end

function gen_cache_init_code(machine::Machine)
    initializeCaches = Vector{Expr}()

    for device in machine.devices
        push!(initializeCaches, gen_cache_init_code(device))
    end

    return Expr(:block, initializeCaches...)
end

function gen_input_assignment_code(
    inputSymbols::Dict{String, Vector{Symbol}},
    processDescription::AbstractProcessDescription,
    machine::Machine,
    processInputSymbol::Symbol = :input,
)
    @assert length(inputSymbols) >=
            sum(values(in_particles(processDescription))) + sum(values(out_particles(processDescription))) "Number of input Symbols is smaller than the number of particles in the process description"

    assignInputs = Vector{Expr}()
    for (name, symbols) in inputSymbols
        type = type_from_name(name)
        index = parse(Int, name[2:end])

        p = nothing

        if (index > in_particles(processDescription)[type])
            index -= in_particles(processDescription)[type]
            @assert index <= out_particles(processDescription)[type] "Too few particles of type $type in input particles for this process"

            p = "filter(x -> typeof(x) <: $type, out_particles($(processInputSymbol)))[$(index)]"
        else
            p = "filter(x -> typeof(x) <: $type, in_particles($(processInputSymbol)))[$(index)]"
        end

        for symbol in symbols
            # TODO generate correct access expression
            # TODO how to define cahce strategies?
            device = machine.devices[1]
            evalExpr = eval(gen_access_expr(device, symbol))
            push!(assignInputs, Meta.parse("$(evalExpr) = ParticleValue($p, 1.0)"))
        end
    end

    return Expr(:block, assignInputs...)
end

"""
    get_compute_function(graph::DAG, process::AbstractProcessDescription, machine::Machine)

Return a function of signature `compute_<id>(input::AbstractProcessInput)`, which will return the result of the DAG computation on the given input.
"""
function get_compute_function(graph::DAG, process::AbstractProcessDescription, machine::Machine)
    (code, inputSymbols, outputSymbol) = gen_code(graph, machine)

    initCaches = gen_cache_init_code(machine)
    assignInputs = gen_input_assignment_code(inputSymbols, process, machine, :input)

    # TODO generate correct access expression
    # TODO how to define cache strategies?
    device = machine.devices[1]

    functionId = to_var_name(UUIDs.uuid1(rng[1]))
    resSym = eval(gen_access_expr(device, outputSymbol))
    expr = Meta.parse(
        "function compute_$(functionId)(input::AbstractProcessInput) $initCaches; $assignInputs; $code; return $resSym; end",
    )
    func = eval(expr)

    return func
end

"""
    execute(graph::DAG, process::AbstractProcessDescription, machine::Machine, input::AbstractProcessInput)

Execute the code of the given `graph` on the given input particles.

This is essentially shorthand for
    ```julia
    compute_graph = get_compute_function(graph, process)
    result = compute_graph(particles)
    ```

See also: [`parse_dag`](@ref), [`parse_process`](@ref), [`gen_process_input`](@ref)
"""
function execute(graph::DAG, process::AbstractProcessDescription, machine::Machine, input::AbstractProcessInput)
    (code, inputSymbols, outputSymbol) = gen_code(graph, machine)

    initCaches = gen_cache_init_code(machine)
    assignInputs = gen_input_assignment_code(inputSymbols, process, machine, :input)

    # TODO generate correct access expression
    # TODO how to define cache strategies?
    device = machine.devices[1]

    functionId = to_var_name(UUIDs.uuid1(rng[1]))
    resSym = eval(gen_access_expr(device, outputSymbol))
    expr = Meta.parse(
        "function compute_$(functionId)(input::AbstractProcessInput) $initCaches; $assignInputs; $code; return $resSym; end",
    )
    func = eval(expr)

    result = 0
    try
        result = @eval $func($input)
    catch e
        println("Error while evaluating: $e")

        println("Function:\n$expr")
        @assert false
    end

    return result
end
