
"""
    compute(t::FusedComputeTask, data)

Compute a [`FusedComputeTask`](@ref). This simply asserts false and should not be called. Fused Compute Tasks generate their expressions directly through the other tasks instead.
"""
function compute(t::FusedComputeTask, data)
    @assert false "This is not implemented and should never be called"
end

"""
    get_expression(t::FusedComputeTask, device::AbstractDevice, inExprs::Vector{String}, outExpr::String)

Generate code evaluating a [`FusedComputeTask`](@ref) on `inExprs`, providing the output on `outExpr`.
`inExprs` should be of the correct types and may be heterogeneous. `outExpr` will be of the type of the output of `T2` of t.
"""
function get_expression(t::FusedComputeTask, device::AbstractDevice, inExprs::Vector, outExpr)
    c1 = length(t.t1_inputs)
    c2 = length(t.t2_inputs) + 1
    expr1 = nothing
    expr2 = nothing

    cacheStrategy = cache_strategy(device)

    inExprs1 = Vector()
    for sym in t.t1_inputs
        push!(inExprs1, gen_access_expr(device, sym))
    end

    outExpr1 = gen_access_expr(device, t.t1_output)

    inExprs2 = Vector()
    for sym in t.t2_inputs
        push!(inExprs2, gen_access_expr(device, sym))
    end

    expr1 = get_expression(t.first_task, device, inExprs1, outExpr1)
    expr2 = get_expression(t.second_task, device, [inExprs2..., outExpr1], outExpr)

    full_expr = Expr(:block, expr1, expr2)

    return full_expr
end

"""
    get_expression(node::ComputeTaskNode, device::AbstractDevice)

Generate and return code for a given [`ComputeTaskNode`](@ref).
"""
function get_expression(node::ComputeTaskNode, device::AbstractDevice)
    t = typeof(node.task)
    @assert length(node.children) <= children(node.task) "Node $(node) has too many children for its task: node has $(length(node.children)) versus task has $(children(node.task))\nNode's children: $(getfield.(node.children, :children))"

    # TODO get device from the node
    cacheStrategy = cache_strategy(device)

    inExprs = Vector()
    for id in getfield.(node.children, :id)
        push!(inExprs, gen_access_expr(device, Symbol(to_var_name(id))))
    end
    outExpr = gen_access_expr(device, Symbol(to_var_name(node.id)))

    return get_expression(node.task, device, inExprs, outExpr)
end

"""
    get_expression(node::DataTaskNode, device::AbstractDevice)

Generate and return code for a given [`DataTaskNode`](@ref).
"""
function get_expression(node::DataTaskNode, device::AbstractDevice)
    @assert length(node.children) <= 1

    # TODO: do things to transport data from/to gpu, between numa nodes, etc.
    # TODO get device from the node

    cacheStrategy = cache_strategy(device)
    inExpr = nothing
    if (length(node.children) == 1)
        inExpr = eval(gen_access_expr(device, Symbol(to_var_name(node.children[1].id))))
    else
        inExpr = eval(gen_access_expr(device, Symbol("$(to_var_name(node.id))_in")))
    end
    outExpr = eval(gen_access_expr(device, Symbol(to_var_name(node.id))))

    dataTransportExp = Meta.parse("$outExpr = $inExpr")

    return dataTransportExp
end
